#include <iostream.h>
#include <windows.h>
#include <fstream.h>

#include <stdio.h>

const int MAX_DATA_SIZE = 1024;
const int MAX_LINE_SIZE = 1024;

#define MIDC 60

// a packet looks like this:
// ---------------------------------
// | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 |
// ---------------------------------
//
// ---------------------------------
// | 0 |		  TEMPO            |
// ---------------------------------
//
// ---------------------------------
// | 1 |	      TNE	           |
// |ID1    |     NOTE1             |
// | NOTE1 |   ID2 |    NOTE2      |
// |     NOTE2     |  ID3  | NOTE3 |
// |         NOTE3         |  ID4  |
// |              NOTE4            |
// ---------------------------------
typedef struct packet{
		unsigned note4:8;
		unsigned id4:2;
		unsigned note3:8;
		unsigned id3:2;
		unsigned note2:8;
		unsigned id2:2;
		unsigned note1:8;
		unsigned id1:2;
		unsigned tne_tempo:7;
		unsigned id:1;
} packet;

// we are going to have a linked list of packet nodes
typedef struct _cmd {
	packet p;
	_cmd *next;
}_cmd;

/////////////////////////////////////////////////////////////
// helper functions
void setnotes(_cmd *node, char *delims);
unsigned char getnote(char *note, char octave);
void printlist(_cmd *head);
/////////////////////////////////////////////////////////////


void main(int argc, char* argv[]){
	// make sure we have an argument
	if(argc < 2){
		cout << "Please specify a music file as input" << endl;
		return;
	}
	
	//////////////////////////////////////////////////////////////////////////
	// Handle to the serial port
	HANDLE hCom;
	
	// open the serial port (COM1)
	cout << "Opening serial port, COM1 . . . ";
	hCom = CreateFile("COM1:", GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, 0, 0);
	if(hCom == INVALID_HANDLE_VALUE){
		cout << "Error opening COM1" << endl;
		return;
	}else{
		cout << "DONE" << endl << endl;
	}

	// we also have to configure the serial port before using it 
	DCB PortDCB;

	PortDCB.DCBlength = sizeof(DCB);
	GetCommState(hCom, &PortDCB);
	
	PortDCB.BaudRate = 2400;
	PortDCB.fBinary = true;
	PortDCB.fParity = false;
	PortDCB.fOutxCtsFlow = false;
	PortDCB.fOutxDsrFlow = false;
	PortDCB.fDtrControl = DTR_CONTROL_ENABLE;
	PortDCB.fDsrSensitivity = false;
	PortDCB.fTXContinueOnXoff = false;
	PortDCB.fOutX = false;
	PortDCB.fInX = false;
	PortDCB.fErrorChar = false;
	PortDCB.fNull = false;
	PortDCB.fRtsControl = RTS_CONTROL_ENABLE;
	PortDCB.fAbortOnError = false;
	PortDCB.ByteSize = 8;
	PortDCB.Parity = NOPARITY;
	PortDCB.StopBits = ONESTOPBIT;

	if(!SetCommState(hCom, &PortDCB)){
		cout << "Couldn't configure the serial port!" << endl;
		return;
	}

	
	// open and parse the music file
	DWORD bytesread, byteswritten;
	ifstream fmusic(argv[1], ios::in | ios::nocreate, filebuf::openprot);
	char line[MAX_LINE_SIZE];
	char *token;
	char *delims = "\n\t; ";
	_cmd *head = NULL;
	_cmd *tail = NULL;
	unsigned char cval;

	// get each line, parse into tokens and create packet to send to 8051
	while(fmusic.getline(line, MAX_LINE_SIZE, '\n')){
		token = strtok(line, delims);
		if(line[0] != '#' && token != NULL){
			_cmd *node = new _cmd;
			node->next = NULL;
			
			while(token != NULL){
				//cout << token << " ";
				if(!strcmp(token, "tempo")){
					node->p.id = 0;
					token = strtok(NULL, delims);
					cval = (unsigned char)atoi(token);
					if(cval > 9) cval = 9;
					else if(cval < 0) cval = 0;
					node->p.tne_tempo = cval;
				}else if(!strcmp(token, "tne")){
					node->p.id = 1;
					token = strtok(NULL, delims);
					cval = (unsigned char)atoi(token);
					node->p.tne_tempo = cval;

					setnotes(node, delims);
				}else{
					cout << "Error! Invalid keword: " << token << endl;
					return;
				}
				token = strtok(NULL, delims);
			}
		
			if(head == NULL){
				head = tail = node;
			}else{
				tail->next = node;
				tail = node;
			}
		}
		//cout << endl;
	}
	fmusic.close();

//	printlist(head);

	_cmd *curr = head;

	//==============================================================================
	// this is the code to wait on the serial port (wait for 8051 to send us an ack)
	DWORD eMask;
	unsigned char chRead;
	if(!SetCommMask(hCom, EV_RXCHAR)){//hCom, EV_RXCHAR | EV_CTS | EV_DSR | EV_ERR | EV_RING | EV_RLSD | EV_RXCHAR | EV_RXFLAG | EV_TXEMPTY)){
		cout << "Error setting CommMask!!" << endl;
		return;
	}
	
	unsigned char sendoff;
	int packetcount = 0;
	// send byte for byte until traverse the entire list
	while(curr != NULL){
		if(curr->p.id)
			packetcount = 6;
		else
			packetcount = 1;

		for(int i = 0; i < packetcount; i++){
			sendoff = 0x00;
			// create the packet
			switch(i){
			case 0:
				sendoff = curr->p.id << 7;
				sendoff = sendoff | (curr->p.tne_tempo & 0x7f);
				break;
			case 1:
				sendoff = curr->p.id1 << 6;
				sendoff = sendoff | (curr->p.note1 >> 2);
				break;
			case 2:
				sendoff = curr->p.note1 << 6;
				sendoff = sendoff | (curr->p.id2 << 4) | (curr->p.note2 >> 4);
				break;
			case 3:
				sendoff = curr->p.note2 << 4;
				sendoff = sendoff | (curr->p.id3 << 2) | (curr->p.note3 >> 6);
				break;
			case 4:
				sendoff = curr->p.note3 << 2;
				sendoff = sendoff | curr->p.id4;
				break;
			case 5:
				sendoff = curr->p.note4;
				break;
			default:
				break;
			}
			cout << "Sending data. . ." << endl;
			// always send the first byte
			printf("Sending: 0x%x\n", sendoff);
			if(!WriteFile(hCom, &sendoff, 1, &byteswritten, NULL)){
				LPVOID lpMsgBuf;
				FormatMessage( 
					FORMAT_MESSAGE_ALLOCATE_BUFFER | 
					FORMAT_MESSAGE_FROM_SYSTEM | 
					FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL,
					GetLastError(),
					0, // Default language
					(LPTSTR) &lpMsgBuf,
					0,
					NULL 
				);
				// Display the string.
				MessageBox( NULL, (LPCTSTR)lpMsgBuf, "Error", MB_OK | MB_ICONINFORMATION );
				// Free the buffer.
				LocalFree( lpMsgBuf );

				cout << "Error writing to serial port!" << endl;
				exit(0);
			}
			cout << "Waiting for ack . . ." << endl;
			if(WaitCommEvent(hCom, &eMask, NULL)){
				if(ReadFile(hCom, &chRead, 1, &bytesread, NULL)){
					cout << "Character read: " << chRead << endl;
					printf("Received: 0x%x\n", chRead);
				}else{
					cout << "Error reading COM port!" << endl;
					exit(0);
				}
				if(chRead != 'A'){
					cout << "Didn't receive a proper acknowledgement!!" << endl;
					exit(0);
				}

			}else{
				cout << "Error waiting for CommEvent!!" << endl;
				exit(0);
			}
		}

		curr = curr->next;
	}
	//==============================================================================

	// free the memory
	curr = head;
	while(curr != NULL){
		curr = curr->next;
		delete head;
		head = curr;
	}

	cout << "finished" << endl;

	// close the files
	CloseHandle(hCom);
}

// parses the current line and sets the values of the notes
void setnotes(_cmd *node, char *delims){
	char *token = strtok(NULL, delims);
	char *token2;
	unsigned char note;

	// initialize all voices to 0 (no voice)
	node->p.id1 = 0;
	node->p.id2 = 0;
	node->p.id3 = 0;
	node->p.id4 = 0;
	
	if(token == NULL){
		return;
	}
	//NOTE1//////////////////////////////////////////////////
	// set the voice 
	switch(token[0]){
	case 'p':
		node->p.id1 = 1;
		break;
	case 'v':
		node->p.id1 = 2;
		break;
	case 't':
		node->p.id1 = 3;
		break;
	case 'n':
		break;
	default:
		cout << "Error! Invalid voice type: " << token[0];
		exit(0);
	}
	// set the value of the note
	// token holds the note, and token2 holds the octave
	token = strtok(NULL, delims);
	token2 = strtok(NULL, delims);
	if(token2 == NULL){
		cout << "Error! Invalid note specification. Make sure the notes and octaves are on the same line.";
		exit(0);
	}
	note = getnote(token, (char)atoi(token2));
	node->p.note1 = note;

	token = strtok(NULL, delims);

	if(token == NULL){
		return;
	}
	//NOTE2//////////////////////////////////////////////////
	// set the voice 
	switch(token[0]){
	case 'p':
		node->p.id2 = 1;
		break;
	case 'v':
		node->p.id2 = 2;
		break;
	case 't':
		node->p.id2 = 3;
		break;
	case 'n':
		break;
	default:
		cout << "Error! Invalid voice type: " << token[0];
		exit(0);
	}
	// set the value of the note
	token = strtok(NULL, delims);
	token2 = strtok(NULL, delims);
	if(token2 == NULL){
		cout << "Error! Invalid note specification. Make sure the notes and octaves are on the same line.";
		exit(0);
	}
	note = getnote(token, (char)atoi(token2));
	node->p.note2 = note;

	token = strtok(NULL, delims);

	if(token == NULL){
		return;
	}

	//NOTE3//////////////////////////////////////////////////
	// set the voice 
	switch(token[0]){
	case 'p':
		node->p.id3 = 1;
		break;
	case 'v':
		node->p.id3 = 2;
		break;
	case 't':
		node->p.id3 = 3;
		break;
	case 'n':
		break;
	default:
		cout << "Error! Invalid voice type: " << token[0];
		exit(0);
	}
	// set the value of the note
	// token holds the note, and token2 holds the octave
	token = strtok(NULL, delims);
	token2 = strtok(NULL, delims);
	if(token2 == NULL){
		cout << "Error! Invalid note specification. Make sure the notes and octaves are on the same line.";
		exit(0);
	}
	note = getnote(token, (char)atoi(token2));
	node->p.note3 = note;

	token = strtok(NULL, delims);
	
	if(token == NULL){
		return;
	}

	//NOTE4//////////////////////////////////////////////////
	// set the voice 
	switch(token[0]){
	case 'p':
		node->p.id4 = 1;
		break;
	case 'v':
		node->p.id4 = 2;
		break;
	case 't':
		node->p.id4 = 3;
		break;
	case 'n':
		break;
	default:
		cout << "Error! Invalid voice type: " << token[0];
		exit(0);
	}
	// set the value of the note
	// token holds the note, and token2 holds the octave
	token = strtok(NULL, delims);
	token2 = strtok(NULL, delims);
	if(token2 == NULL){
		cout << "Error! Invalid note specification. Make sure the notes and octaves are on the same line.";
		exit(0);
	}
	note = getnote(token, (char)atoi(token2));
	node->p.note4 = note;
}

// converts a note and an octave to a value between 0-127
unsigned char getnote(char *note, char octave){
	unsigned char noteval = MIDC;

	switch(note[0]){
	case 'a':
		noteval -= 3;
		break;
	case 'b':
		noteval -= 1;
		break;
	case 'c':
		break;
	case 'd':
		noteval += 2;
		break;
	case 'e':
		noteval += 4;
		break;
	case 'f':
		noteval += 5;
		break;
	case 'g':
		noteval += 7;
		break;
	default:
		cout << "Error! Invalid note: " << note;
		exit(0);
		break;
	}

	// check to see if we have sharps or flats
	if(strlen(note) > 1){
		if(note[1] == '#')
			noteval++;
		else if(note[1] == 'f')
			noteval--;
		else{
			cout << "Error! Invalid note: " << note;
			exit(0);
		}
	}

	// set the octave relative to middle C
	noteval = noteval + (12 * octave);

	return noteval;
}

// print each node in the linked list
void printlist(_cmd *head){
	_cmd *curr = head;
	while(curr != NULL){
		if(curr->p.id){
			cout << "Read a TNE: " << (int)curr->p.tne_tempo << endl;
			if(curr->p.id1 == 0)
				cout << "no voice " << endl;
			else if(curr->p.id1 == 1)
				cout << "piano " << (int)curr->p.note1 << endl;
			else if(curr->p.id1 == 2)
				cout << "violin " << (int)curr->p.note1 << endl;
			else if(curr->p.id1 == 3)
				cout << "trumpet " << (int)curr->p.note1 << endl;
			else cout << (int)curr->p.id1 << endl;

			if(curr->p.id2 == 0)
				cout << "no voice" << endl;
			else if(curr->p.id2 == 1)
				cout << "piano " << (int)curr->p.note2 << endl;
			else if(curr->p.id2 == 2)
				cout << "violin " << (int)curr->p.note2 << endl;
			else if(curr->p.id2 == 3)
				cout << "trumpet " << (int)curr->p.note2 << endl;
			else cout << (int)curr->p.id2 << endl;

			if(curr->p.id3 == 0)
				cout << "no voice" << endl;
			else if(curr->p.id3 == 1)
				cout << "piano " << (int)curr->p.note3 << endl;
			else if(curr->p.id3 == 2)
				cout << "violin " << (int)curr->p.note3 << endl;
			else if(curr->p.id3 == 3)
				cout << "trumpet " << (int)curr->p.note3 << endl;
			else cout << (int)curr->p.id3 << endl;

			if(curr->p.id4 == 0)
				cout << "no voice" << endl;
			else if(curr->p.id4 == 1)
				cout << "piano " << (int)curr->p.note4 << endl;
			else if(curr->p.id4 == 2)
				cout << "violin " << (int)curr->p.note4 << endl;
			else if(curr->p.id4 == 3)
				cout << "trumpet " << (int)curr->p.note4 << endl;
			else cout << (int)curr->p.id4 << endl;
		}else{
			cout << "Read a tempo: value = " << (int)curr->p.tne_tempo << endl;;			
		}
		curr = curr->next;
	}
}